/* smandel.c
 * ------
 * Simple mandelbrot calculation; creates a jpeg image
 *
 * The OMPi team, 2014.
 */

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>
#include <math.h>
#include <jpeglib.h>
#include <string.h>

#define IMAGE_HEIGHT 2160
#define IMAGE_WIDTH  3840
#define COLOR_COMPONENTS 3
#define MAX_ITER 64

#define IMGFILE "mandelbrot_ppg.jpg"

JSAMPLE image_buffer[IMAGE_HEIGHT*IMAGE_WIDTH*COLOR_COMPONENTS];


GLOBAL(void)
write_JPEG_file (char * filename, int quality)
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	FILE * outfile;
	JSAMPROW row_pointer[1];
	int row_stride;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);

	if ((outfile = fopen(filename, "wb")) == NULL) {
		fprintf(stderr, "can't open %s\n", filename);
		exit(1);
	}
	jpeg_stdio_dest(&cinfo, outfile);

	cinfo.image_width = IMAGE_WIDTH;
	cinfo.image_height = IMAGE_HEIGHT;
	cinfo.input_components = COLOR_COMPONENTS;
	cinfo.in_color_space = JCS_RGB;
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE);
	jpeg_start_compress(&cinfo, TRUE);
	row_stride = IMAGE_WIDTH * COLOR_COMPONENTS;

	while (cinfo.next_scanline < cinfo.image_height) {
		row_pointer[0] = & image_buffer[cinfo.next_scanline * row_stride];
		(void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}

	jpeg_finish_compress(&cinfo);
	fclose(outfile);
	jpeg_destroy_compress(&cinfo);
}

int main(void)
{
	double t1, t2;
	int executed_on_host;

	printf("\n\nCalculating Mandelbrot set for an %d x %d image..\n", IMAGE_WIDTH, IMAGE_HEIGHT);
	
	t1 = omp_get_wtime();
	#pragma omp target data map(from:image_buffer)
	{
		#pragma omp parallel num_threads( omp_get_default_device() == 0 ? 2 : 16 )
		{
			int my_id = omp_get_thread_num();
			int start = (IMAGE_HEIGHT/omp_get_num_threads()) * my_id;
			int end   = (IMAGE_HEIGHT/omp_get_num_threads()) * (my_id + 1);
			
			#pragma omp target map(to:start, end) map(from:executed_on_host)
			{
				int r, c, p, iteration, color;
				float x, y, x0, y0, xtemp;
				const float downscaley = 2.0/IMAGE_HEIGHT;
				const float downscalex = 3.5/IMAGE_WIDTH;

				executed_on_host = omp_is_initial_device();

				for(r=start; r<end; r++)
				{
					for(c=0; c<IMAGE_WIDTH; c++)
					{
						x0 = c * downscalex - 2.5;
						y0 = r * downscaley - 1.0;

						x = 0.0;
						y = 0.0;

						for(iteration = 0; x*x + y*y <= (2*2)  &&  iteration < MAX_ITER; iteration ++)
						{
							xtemp = x*x - y*y + x0;
							y = 2*x*y + y0;
							x = xtemp;
						}

						color = (iteration >= MAX_ITER) ? 0.0f : iteration*iteration*iteration*iteration;
						image_buffer[(r*IMAGE_WIDTH + c)*COLOR_COMPONENTS + 0] = ((char *)&color)[2];
						image_buffer[(r*IMAGE_WIDTH + c)*COLOR_COMPONENTS + 1] = ((char *)&color)[1];
						image_buffer[(r*IMAGE_WIDTH + c)*COLOR_COMPONENTS + 2] = ((char *)&color)[0];
					}
				}
			}
		}
	}
	t2 = omp_get_wtime();
	
	write_JPEG_file(IMGFILE, 100);
	printf("Calculation ok!\nOutput file is %s\nTime elapsed=%.8f (@ %s)\n", IMGFILE, (double)(t2-t1),
	       (executed_on_host == 1) ? "Zynq": "Epiphany");
	
	return 0;
}
