/* pi.c
 * ----
 * Simple pi calculation targetted to the epiphany, using OMPi.
 *
 * The OMPi team, 2014.
 */

#include <stdio.h>
#include <omp.h>

#define N 2000000      /* Intervals */


/* Naive pi calculation based on the trapezoid rule.
 */
void calc_pi(int nthr)
{
	double W = 4.0 / N, W2 = W*W/16.0, pi = 0.0;
	
	/* Executed @ host (zynq, aka device 0) */
	#pragma omp parallel num_threads(nthr) reduction(+:pi)
	{
		int i = omp_get_thread_num();
		
		/* Offloaded @ default device (normally epiphany). 
		 * Note that pi is implicitly mapped as tofrom.
		 */
		#pragma omp target map(to:i,W,W2,nthr)
		{
			for ( ; i < N; i += nthr)
				pi += W / (1.0 + (0.5 + i)*(0.5 + i)*W2);
			
		}
	}
	printf("pi = %.10lf", pi);
}


/* Calculate pi with 1, 2, 4, 8, 16 threads/kernels. 
 */
void main()
{
	int    nthr;
	double t1, t2;
	
	printf("Calculating pi on the %s, using 1, 2, 4, 8, and 16 kernels..\n",
	       omp_get_default_device() == 0 ? "Zynq" : "Epiphany");
	for (nthr = 1; nthr <= 16; nthr <<= 1)
	{
		t1 = omp_get_wtime();
		calc_pi(nthr);
		t2 = omp_get_wtime();
		
		printf("\t[%2d threads => %2.5lf sec]\n", nthr, t2-t1);
	}
}

